/*
 * Priscila: A GNU/EDMA OpenGL Presentation Tool
 * Copyright (C) 2004  David Martnez Oliveira
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.

 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 

/*
 * GL_ITEM_IMAGE Class
 * Encapsulates a image presentation item object
 * -------------------------------------------------------------
 * September, 5th, 2004
 * - File Creation. GNU/EDMA in-line class
 * ---------------------------------------------------------------
 * September, 8th, 2004
 * - Adding save/load methods
 * ----------------------------------------------------------------
 * September, 9th, 2004
 * - Finished load method
 */

/* OpenGL includes */
#include <GL/gl.h>
#include <GL/glu.h>

/* SDL and SDL TrueType Font include */
#include <SDL.h>

#include <edma.h>

#include "sdlgl_tex.h"
#include "gl_item_text.h"


/* Data structures */
/* -- Data structure for GL_ITEM_IMAGE */
typedef struct
{
  EPChar      data;
  EPVoid      handler;
} GL_ITEM_IMAGEData;


typedef struct
{
  EReal64    w, h;
  EReal64    r, g, b, alpha;
} GL_ITEM_QUADData;

/* -- Method prototypes for GL_ITEM_IMAGE Class*/

ESint32 EDMAPROC GL_ITEM_IMAGEcreate (OBJID, EPChar);
ESint32 EDMAPROC GL_ITEM_IMAGEdestroy (OBJID);
ESint32 EDMAPROC GL_ITEM_IMAGErip (OBJID);
ESint32 EDMAPROC GL_ITEM_IMAGErender (OBJID);
ESint32 EDMAPROC GL_ITEM_IMAGEsave (OBJID, OBJID);
ESint32 EDMAPROC GL_ITEM_IMAGEload (OBJID, OBJID);


/* -- Method prototypes for GL_ITEM_QUAD Class*/

ESint32 EDMAPROC GL_ITEM_QUADcreate (OBJID, EPChar);
ESint32 EDMAPROC GL_ITEM_QUADdestroy (OBJID);
ESint32 EDMAPROC GL_ITEM_QUADrip (OBJID);
ESint32 EDMAPROC GL_ITEM_QUADrender (OBJID);
ESint32 EDMAPROC GL_ITEM_QUADsave (OBJID, OBJID);
ESint32 EDMAPROC GL_ITEM_QUADload (OBJID, OBJID);



ESint32
gl_item_image_class_factory ()
{
  CLASSID    cid;

  cid = edma_idf_get_free_class_id (EDMA_LOCAL_CLASS);

  edma_idf_set_class_name (cid, "GL_ITEM_IMAGE");
  edma_idf_set_class_namespace (cid, "PRISCILA");
  edma_idf_set_class_version (cid, 0, 0);

  /* Item position in 3D space*/
  edma_add_local_class_property (cid, "data", DT_EZSTRING, E_L, 0);
  edma_add_local_class_property (cid, "handler", DT_EPOINTER, E_L, 0);

  /* GL_ITEM methods */
  edma_add_local_class_method (cid, "create", "OZ", (PPROC) GL_ITEM_IMAGEcreate, 0, 0, 0);
  edma_add_local_class_method (cid, "destroy", "", (PPROC) GL_ITEM_IMAGEdestroy, 0, 0, 0);
  edma_add_local_class_method (cid, "rip", "", (PPROC) GL_ITEM_IMAGErip, 0, 0, 0);


  edma_add_local_class_method (cid, "render", "", (PPROC) GL_ITEM_IMAGErender, 0, 0, 0);
  edma_add_local_class_method (cid, "save", "O", (PPROC) GL_ITEM_IMAGEsave, 0, 0, 0);
  edma_add_local_class_method (cid, "load", "O", (PPROC) GL_ITEM_IMAGEload, 0, 0, 0);

  /* GL_ITEM_TEXT inherits from GL_ITEM */
  edma_add_local_class_superclass_by_name (cid, "GL_ITEM", "SUPER", "INNER");

  //edma_local_class_finish (cid);
  edma_idf_set_class_id (cid);


  return 0;
}




/* -- Method prototypes for GL_ITEM_TEXT Class*/

ESint32 EDMAPROC 
GL_ITEM_IMAGEcreate (OBJID obj, EPChar str)
{
  double            scale;
  GL_ITEM_IMAGEData *m = edma_get_data_ref (obj);

  if (str == NULL) return -1;

  m->handler = (void *) SDLGL_new_texture (); 
  if (m->handler != NULL)
    {
      SDLGL_read_from_file (m->handler, str);
      SDLGL_tex_set_scale (m->handler, 1.0f);
      edma_wprop3 (obj, "data", str);
      /* FIXME: Rework this operation mode for aspect ratio correction*/
      
    }
  else
    return -1;


  return 0;
}


ESint32 EDMAPROC 
GL_ITEM_IMAGEdestroy (OBJID obj)
{
  GL_ITEM_IMAGEData *m = edma_get_data_ref (obj);

  SDLGL_free_texture (m->handler);

  return 0;
}


ESint32 EDMAPROC 
GL_ITEM_IMAGErip (OBJID obj)
{
  GL_ITEM_IMAGEData *m = edma_get_data_ref (obj);

  SDLGL_free_texture (m->handler);

  return 0;
}


ESint32 EDMAPROC 
GL_ITEM_IMAGErender (OBJID obj)
{
  double   alpha;
  GL_ITEM_IMAGEData *m = edma_get_data_ref (obj);
  
  edma_met3 (obj, "SUPER>render");
  edma_rprop3 (obj, "SUPER>alpha", &alpha);
  
  /* If image is not transparent disable blending */
  if (alpha == 1.0)
    glDisable(GL_BLEND);
  
  //glEnable(GL_DEPTH_TEST);

  SDLGL_img_render (m->handler);
  
  //if (alpha == 1.0)
  //glDisable(GL_BLEND);

  //glDisable(GL_DEPTH_TEST);

  return 0;
}


ESint32 EDMAPROC 
GL_ITEM_IMAGEsave (OBJID obj, OBJID stream)
{
  GL_ITEM_IMAGEData *m = edma_get_data_ref (obj);

  edma_met3 (stream, "puts", "GL_ITEM_IMAGE\n");

  /* Let superclass save common information */
  edma_met3 (obj, "SUPER>save", stream);

  edma_met3 (stream, "puts", m->data);
  edma_met3 (stream, "puts", "\n");



  return 0;
}

/* Load method is "INNER" called from superclass GL_ITEM */
ESint32 EDMAPROC 
GL_ITEM_IMAGEload (OBJID obj, OBJID stream)
{
  EChar            buffer[1024];
  ESint32          len;
  GL_ITEM_IMAGEData *m = edma_get_data_ref (obj);

  len = 1024;
  edma_met3 (stream, "gets", buffer, &len);
  buffer[strlen(buffer) - 1] = 0; /* Chomp String */
  
  edma_wprop3 (obj, "data", buffer);
  edma_met3 (obj, "create", buffer);

  return 0;
}

/**********************************************************************************/
/**********************************************************************************/
/*  QUAD Item definition */
ESint32
gl_item_quad_class_factory ()
{
  CLASSID    cid;

  cid = edma_idf_get_free_class_id (EDMA_LOCAL_CLASS);

  edma_idf_set_class_name (cid, "GL_ITEM_QUAD");
  edma_idf_set_class_namespace (cid, "PRISCILA");
  edma_idf_set_class_version (cid, 0, 0);

  /* Item position in 3D space*/
  edma_add_local_class_property (cid, "w", DT_EREAL64, E_L, 0);
  edma_add_local_class_property (cid, "h", DT_EREAL64, E_L, 0);
  edma_add_local_class_property (cid, "r", DT_EREAL64, E_L, 0);
  edma_add_local_class_property (cid, "g", DT_EREAL64, E_L, 0);
  edma_add_local_class_property (cid, "b", DT_EREAL64, E_L, 0);
  edma_add_local_class_property (cid, "alpha", DT_EREAL64, E_L, 0);

  /* GL_ITEM methods */
  edma_add_local_class_method (cid, "create", "OZ", (PPROC) GL_ITEM_QUADcreate, 0, 0, 0);
  edma_add_local_class_method (cid, "destroy", "", (PPROC) GL_ITEM_QUADdestroy, 0, 0, 0);
  edma_add_local_class_method (cid, "rip", "", (PPROC) GL_ITEM_QUADrip, 0, 0, 0);


  edma_add_local_class_method (cid, "render", "", (PPROC) GL_ITEM_QUADrender, 0, 0, 0);
  edma_add_local_class_method (cid, "save", "O", (PPROC) GL_ITEM_QUADsave, 0, 0, 0);
  edma_add_local_class_method (cid, "load", "O", (PPROC) GL_ITEM_QUADload, 0, 0, 0);

  /* GL_ITEM_TEXT inherits from GL_ITEM */
  edma_add_local_class_superclass_by_name (cid, "GL_ITEM", "SUPER", "INNER");

  edma_idf_set_class_id (cid);


  return 0;
}




/* -- Method prototypes for GL_ITEM_TEXT Class*/

ESint32 EDMAPROC 
GL_ITEM_QUADcreate (OBJID obj, EPChar str)
{
  double            scale;
  GL_ITEM_QUADData *m = edma_get_data_ref (obj);

  /* Nothing to do with a quad */

  return 0;
}


ESint32 EDMAPROC 
GL_ITEM_QUADdestroy (OBJID obj)
{
  GL_ITEM_QUADData *m = edma_get_data_ref (obj);

  /* Nothing to do with a quad */

  return 0;
}


ESint32 EDMAPROC 
GL_ITEM_QUADrip (OBJID obj)
{
  GL_ITEM_QUADData *m = edma_get_data_ref (obj);

  /* Nothing to do with a quad */

  return 0;
}


ESint32 EDMAPROC 
GL_ITEM_QUADrender (OBJID obj)
{
  GL_ITEM_QUADData *m = edma_get_data_ref (obj);
  
  edma_met3 (obj, "SUPER>render");

  glBindTexture (GL_TEXTURE_2D, 0);
  glBegin(GL_QUADS);
  glNormal3f (0.0f , 0.0f , 0.1f);

      glVertex3f(  m->w,  m->h,  0.0f);
      glVertex3f(  m->w, 0.0f,  0.0f);
      glVertex3f( 0.0f, 0.0f,  0.0f);
      glVertex3f( 0.0f, m->h,  0.0f);
  glEnd();

  return 0;
}


ESint32 EDMAPROC 
GL_ITEM_QUADsave (OBJID obj, OBJID stream)
{
  EChar             buffer[80];
  GL_ITEM_QUADData *m = edma_get_data_ref (obj);

  edma_met3 (stream, "puts", "GL_ITEM_QUAD\n");

  /* Let superclass save common information */
  edma_met3 (obj, "SUPER>save", stream);

  snprintf (buffer, 80, "%lf % lf", m->w, m->h);
  edma_met3 (stream, "puts", buffer);
  edma_met3 (stream, "puts", "\n");



  return 0;
}

/* Load method is "INNER" called from superclass GL_ITEM */
ESint32 EDMAPROC 
GL_ITEM_QUADload (OBJID obj, OBJID stream)
{
  EChar            buffer[1024];
  ESint32          len;
  GL_ITEM_QUADData *m = edma_get_data_ref (obj);

  len = 1024;
  edma_met3 (stream, "gets", buffer, &len);
  buffer[strlen(buffer) - 1] = 0; /* Chomp String */

  sscanf (buffer, "%lf %lf", &m->w, &m->h);
  edma_printf_obj (obj, "Width: %lf Height: %lf", m->w, m->h);

  return 0;
}


