/*
 * Priscila: A GNU/EDMA OpenGL Presentation Tool
 * Copyright (C) 2004  David Martnez Oliveira
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.

 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 

/*
 * GL_ITEM_TEXT Class
 * Encapsulates a text item presentation object and font object
 * -------------------------------------------------------------
 * September, 5th, 2004
 * - File Creation. GNU/EDMA in-line class
 * ----------------------------------------------------------------
 * September, 8th, 2004
 * - Adding save/load methods 
 * ----------------------------------------------------------------
 * September, 16th, 2004
 * - Updated font class to generate a CallList for efficient text
 *   rendering
 * -----------------------------------------------------------------
 * September, 18th, 2004
 * - Added class GL_ITEM_FREE_TEXT for per-character rendering
 */

/* Acknowledges: 
 *   Thanks to NeHe Productions for this excellent tutorials
 *   TTF Font Display list generation is based on NeHe Lesson 43
 *
 *   http://nehe.gamedev.net/data/lessons/lesson.asp?lesson=43
 */


#include <stdlib.h>
/* OpenGL includes */
#include <GL/gl.h>
#include <GL/glu.h>

/* SDL and SDL TrueType Font include */
#include <SDL.h>
#include <SDL_ttf.h>

#include <ft2build.h>
#include <freetype/freetype.h>
#include <freetype/ftglyph.h>

#include <edma.h>

#include "sdlgl_font.h"
#include "gl_item_text.h"


/* Data structures */
typedef struct
{
  EUint32        list_base;
  EUint32        textures[128];
} TEX_FONT;

/* -- Data structures for GL_FONT */
typedef struct 
{
  EPChar       font_id;
  EPChar       font_name;
  ESint32      font_size;
  EReal64      fgcolor_r, fgcolor_g, fgcolor_b;
  EPVoid       handler;
  EDMAT_BUFFER tex;
} GL_FONTData;

/* -- Data structure for GL_ITEM_TEXT */
typedef struct
{
  EPChar      data;
  //OBJID       font;
  EPVoid      handler;
  EPChar      style;
} GL_ITEM_TEXTData;

/* -- Data structure for GL_ITEM_FREE_TEXT */
typedef struct
{
  EPChar      data;
  OBJID       font;
  EPChar      style;
} GL_ITEM_FREE_TEXTData;


/* Method prorotypes goes here */
/* -- Method prototypes for GL_FONT Class */
ESint32 EDMAPROC GL_FONTbuild (OBJID, EPChar, ESint32);
ESint32 EDMAPROC GL_FONTdestroy (OBJID);
ESint32 EDMAPROC GL_FONTrip (OBJID);
ESint32 EDMAPROC GL_FONTset_fgcolor (OBJID, EReal64, EReal64, EReal64);

/* -- Method prototypes for GL_ITEM_TEXT Class*/

ESint32 EDMAPROC GL_ITEM_TEXTcreate (OBJID, OBJID, EPChar);
ESint32 EDMAPROC GL_ITEM_TEXTdestroy (OBJID);
ESint32 EDMAPROC GL_ITEM_TEXTrip (OBJID);
ESint32 EDMAPROC GL_ITEM_TEXTrender (OBJID);
ESint32 EDMAPROC GL_ITEM_TEXTsave (OBJID, OBJID);
ESint32 EDMAPROC GL_ITEM_TEXTload (OBJID, OBJID);

/* -- Method prototypes for GL_ITEM_TEXT Class*/

ESint32 EDMAPROC GL_ITEM_FREE_TEXTcreate (OBJID, OBJID, EPChar);
ESint32 EDMAPROC GL_ITEM_FREE_TEXTdestroy (OBJID);
ESint32 EDMAPROC GL_ITEM_FREE_TEXTrip (OBJID);
ESint32 EDMAPROC GL_ITEM_FREE_TEXTrender (OBJID);
//ESint32 EDMAPROC GL_ITEM_FREE_TEXTrender_txt (OBJID, EPChar, EReal64, EReal64, EPChar);
ESint32 EDMAPROC GL_ITEM_FREE_TEXTrender_txt (CLASSID, EPChar, EReal64, EReal64, EPChar);
ESint32 EDMAPROC GL_ITEM_FREE_TEXTsave (OBJID, OBJID);
ESint32 EDMAPROC GL_ITEM_FREE_TEXTload (OBJID, OBJID);


/* Helper Function Prototypes */
int build_display_list_item ( FT_Face face, char ch, EUint32 list_base, EUint32 *tex_base );

ESint32
gl_font_class_factory ()
{
  CLASSID    cid;

  cid = edma_idf_get_free_class_id (EDMA_LOCAL_CLASS);

  edma_idf_set_class_name (cid, "GL_FONT");
  edma_idf_set_class_namespace (cid, "PRISCILA");
  edma_idf_set_class_version (cid, 0, 0);

  /* Item position in 3D space*/
  edma_add_local_class_property (cid, "font_id", DT_EZSTRING, E_L, 0);
  edma_add_local_class_property (cid, "font_name", DT_EZSTRING, E_L, 0);
  edma_add_local_class_property (cid, "font_size", DT_ESINT32, E_L, 0);

  /* Item color information */
  edma_add_local_class_property (cid, "fg_color_r", DT_EREAL64, E_L, 0);
  edma_add_local_class_property (cid, "fg_color_g", DT_EREAL64, E_L, 0);
  edma_add_local_class_property (cid, "fg_color_b", DT_EREAL64, E_L, 0);

  /* Private values. SDL encapsulation */
  edma_add_local_class_property (cid, "handler", DT_EPOINTER, L, 0);
  edma_add_local_class_property (cid, "tex", DT_EBUFFER, L, 0);


  /* GL_ITEM methods */
  edma_add_local_class_method (cid, "build", "ZS32", (PPROC) GL_FONTbuild, 0, 0, 0);
  edma_add_local_class_method (cid, "destroy", "", (PPROC) GL_FONTdestroy, 0, 0, 0);
  edma_add_local_class_method (cid, "rip", "", (PPROC) GL_FONTrip, 0, 0, 0);

  edma_add_local_class_method (cid, "set_fgcolor", "R32R32R32", (PPROC) GL_FONTset_fgcolor, 0, 0, 0);

  edma_idf_set_class_id (cid);

  return 0;
}



ESint32
gl_item_text_class_factory ()
{
  CLASSID    cid;

  cid = edma_idf_get_free_class_id (EDMA_LOCAL_CLASS);

  edma_idf_set_class_name (cid, "GL_ITEM_TEXT");
  edma_idf_set_class_namespace (cid, "PRISCILA");
  edma_idf_set_class_version (cid, 0, 0);

  /* Item position in 3D space*/
  edma_add_local_class_property (cid, "data", DT_EZSTRING, E_L, 0);
  edma_add_local_class_property (cid, "handler", DT_EPOINTER, E_L, 0);
  edma_add_local_class_property (cid, "style", DT_EZSTRING, E_L, 0);

  /* GL_ITEM methods */
  edma_add_local_class_method (cid, "create", "OZ", (PPROC) GL_ITEM_TEXTcreate, 0, 0, 0);
  edma_add_local_class_method (cid, "destroy", "", (PPROC) GL_ITEM_TEXTdestroy, 0, 0, 0);
  edma_add_local_class_method (cid, "rip", "", (PPROC) GL_ITEM_TEXTrip, 0, 0, 0);


  edma_add_local_class_method (cid, "render", "", (PPROC) GL_ITEM_TEXTrender, 0, 0, 0);
  edma_add_local_class_method (cid, "save", "O", (PPROC) GL_ITEM_TEXTsave, 0, 0, 0);
  edma_add_local_class_method (cid, "load", "O", (PPROC) GL_ITEM_TEXTload, 0, 0, 0);

  /* GL_ITEM_TEXT inherits from GL_ITEM */
  edma_add_local_class_superclass_by_name (cid, "GL_ITEM", "SUPER", "INNER");

  //edma_local_class_finish (cid);
  edma_idf_set_class_id (cid);


  return 0;
}

ESint32
gl_item_free_text_class_factory ()
{
  CLASSID    cid;

  cid = edma_idf_get_free_class_id (EDMA_LOCAL_CLASS);

  edma_idf_set_class_name (cid, "GL_ITEM_FREE_TEXT");
  edma_idf_set_class_namespace (cid, "PRISCILA");
  edma_idf_set_class_version (cid, 0, 0);

  /* Item position in 3D space*/
  edma_add_local_class_property (cid, "data", DT_EZSTRING, E_L, 0);
  edma_add_local_class_property (cid, "font", DT_EOBJECT, E_L, 0);
  edma_add_local_class_property (cid, "style", DT_EZSTRING, E_L, 0);

  /* GL_ITEM methods */
  edma_add_local_class_method (cid, "create", "OZ", (PPROC) GL_ITEM_FREE_TEXTcreate, 0, 0, 0);
  edma_add_local_class_method (cid, "destroy", "", (PPROC) GL_ITEM_FREE_TEXTdestroy, 0, 0, 0);
  edma_add_local_class_method (cid, "rip", "", (PPROC) GL_ITEM_FREE_TEXTrip, 0, 0, 0);


  edma_add_local_class_method (cid, "render", "", (PPROC) GL_ITEM_FREE_TEXTrender, 0, 0, 0);

  /* Static method for independent text rendering */
  edma_add_local_class_method (cid, "render_txt", "ZR64R64Z", (PPROC) GL_ITEM_FREE_TEXTrender_txt, 0, 1, 0);
  edma_add_local_class_method (cid, "save", "O", (PPROC) GL_ITEM_FREE_TEXTsave, 0, 0, 0);
  edma_add_local_class_method (cid, "load", "O", (PPROC) GL_ITEM_FREE_TEXTload, 0, 0, 0);

  /* GL_ITEM_TEXT inherits from GL_ITEM */
  edma_add_local_class_superclass_by_name (cid, "GL_ITEM", "SUPER", "INNER");

  edma_idf_set_class_id (cid);

  return 0;
}

/**** Method implementation ******/

/* Method prorotypes goes here */
/* -- Method prototypes for GL_FONT Class */
ESint32 EDMAPROC 
GL_FONTbuild (OBJID obj, EPChar font_name, ESint32 font_size)
{
  TEX_FONT    *tex;
  FT_Library  library;
  FT_Face     face;
  SDLGL_FONT  the_handler;
  ESint32     i;
  GL_FONTData *m = edma_get_data_ref (obj);

  the_handler = SDLGL_font_new (font_name, font_size);
  if (the_handler != NULL)
    {
      /* Use GNU/EDMA primitives for strings to allow GNU/EDMA to manage memory*/
      edma_wprop3 (obj, "font_name", font_name);
      m->font_size = font_size;
      m->handler = the_handler;
    }
  else
    return -1;

  /* Now Build a Call list for the given font*/
  if ((edma_buffer_alloc (&m->tex, sizeof (TEX_FONT))) < 0)
    {
      edma_printf_obj (obj, "%s", "Cannot allocate memory for GL Calllist data");
      /* FIXME: Clean up memory */
      return -1;
    }

  tex = (TEX_FONT*) m->tex.dat;

  if (FT_Init_FreeType (&library))
    {
      edma_printf_obj (obj, "%s", "Cannot init FT_Library");
      /* FIXME: Clean up memory */
      return -1;
    }

  if (FT_New_Face (library, font_name, 0, &face))
    {
      edma_printf_obj (obj, "%s", "Cannot init FT_Face");
      /* FIXME: Clean up memory*/
      return -1;
    }

  FT_Set_Char_Size (face, font_size << 6, font_size << 6, 96, 96);

  tex->list_base = glGenLists(128);
  glGenTextures( 128, (GLuint*) tex->textures );

  /* Create Calllist*/
  for (i = 0; i < 128; i++)
    build_display_list_item (face, i, tex->list_base, tex->textures);

  FT_Done_Face (face);
  FT_Done_FreeType (library);

  return 0;
}


ESint32 EDMAPROC 
GL_FONTdestroy (OBJID obj)
{
  TEX_FONT    *tex;
  GL_FONTData *m = edma_get_data_ref (obj);

  SDLGL_font_free (m->handler);
  tex = (TEX_FONT*) m->tex.dat;

  glDeleteLists (tex->list_base, 128);
  glDeleteTextures (128, (GLuint*) tex->textures);

  return 0;
}


ESint32 EDMAPROC 
GL_FONTrip (OBJID obj)
{
  GL_FONTData *m = edma_get_data_ref (obj);

  GL_FONTdestroy (obj);

  return 0;
}

ESint32 EDMAPROC 
GL_FONTset_fgcolor (OBJID obj, EReal64 r, EReal64 g, EReal64 b)
{
  GL_FONTData *m = edma_get_data_ref (obj);

  /* Store values for further use */
  m->fgcolor_r = r;
  m->fgcolor_g = g;
  m->fgcolor_b = b;

  /* Denormalize values */
  
  SDLGL_font_set_fgcolor (m->handler, (int)(r * 255), (int) (g * 255), (int)(b * 255));

  return 0;
}


/* -- Method prototypes for GL_ITEM_TEXT Class*/

ESint32 EDMAPROC 
GL_ITEM_TEXTcreate (OBJID obj, OBJID style, EPChar str)
{
  OBJID            styles, font;
  EReal64          scale;
  SDLGL_FONT       font_handler;
  GL_ITEM_TEXTData *m = edma_get_data_ref (obj);


  if (style == -1) return -1;
  if (str == NULL) return -1;

  /* Get singleton and related data */
  styles = (OBJID) edma_smet3 ("SINGLETON", "Instance", "STYLES");
  edma_met3 (styles, "get_data", style, &font, &scale);

  edma_rprop3 (font, "handler", &font_handler);

  m->handler = SDLGL_txt_new (font_handler, str, scale);
  if (m->handler != NULL)
    {
      edma_wprop3 (obj, "data", str);
      edma_wprop3 (obj, "style", style);
    }
  else
    return -1;

  return 0;
}


ESint32 EDMAPROC 
GL_ITEM_TEXTdestroy (OBJID obj)
{
  GL_ITEM_TEXTData *m = edma_get_data_ref (obj);

  SDLGL_txt_free (m->handler);

  return 0;
}


ESint32 EDMAPROC 
GL_ITEM_TEXTrip (OBJID obj)
{
  GL_ITEM_TEXTData *m = edma_get_data_ref (obj);

  SDLGL_txt_free (m->handler);

  return 0;
}


ESint32 EDMAPROC 
GL_ITEM_TEXTrender (OBJID obj)
{
  GL_ITEM_TEXTData *m = edma_get_data_ref (obj);

  edma_met3 (obj, "SUPER>render");

  /*SDLGL_txt_simple_render (m->handler);*/
  SDLGL_txt_simple_render1 (m->handler);

  return 0;
}


ESint32 EDMAPROC 
GL_ITEM_TEXTsave (OBJID obj, OBJID stream)
{
  EChar            buffer[1024];
  GL_ITEM_TEXTData *m = edma_get_data_ref (obj);

  edma_met3 (stream, "puts", "GL_ITEM_TEXT\n");
  /* Let superclass save common information */
  edma_met3 (obj, "SUPER>save", stream);

  /* Save font id */
  edma_met3 (stream, "puts", m->style);
  edma_met3 (stream, "puts", "\n");
  /* Save Text data */
  edma_met3 (stream, "puts", m->data);
  edma_met3 (stream, "puts", "\n");

  return 0;
}

ESint32 EDMAPROC 
GL_ITEM_TEXTload (OBJID obj, OBJID stream)
{
  EChar            style[1024];
  EChar            str[1024];
  ESint32          len;
  GL_ITEM_TEXTData *m = edma_get_data_ref (obj);

  len = 1024;
  edma_met3 (stream, "gets", style, &len);
  style[strlen(style) - 1] = 0; /* Chomp String */

  /* Read text style*/
  len = 1024;
  edma_met3 (stream, "gets", str, &len);
  str[strlen(str) - 1] = 0; /* Chomp String */

  edma_printf_obj (obj, "Style: '%s'  Data: '%s'", style, str);

  /* Create text texture */
  return edma_met3 (obj, "create", style, str);

}



/* -- Method prototypes for GL_ITEM_FREE_TEXT Class*/

ESint32 EDMAPROC 
GL_ITEM_FREE_TEXTcreate (OBJID obj, OBJID style, EPChar str)
{
  OBJID            styles, font;
  EReal64          scale;
  SDLGL_FONT       font_handler;
  GL_ITEM_FREE_TEXTData *m = edma_get_data_ref (obj);


  if (style == -1) return -1;
  if (str == NULL) return -1;

  /* Get singleton and related data */
  styles = (OBJID) edma_smet3 ("SINGLETON", "Instance", "STYLES");
  edma_met3 (styles, "get_data", style, &font, &scale);

  if (font != -1)
    {
      edma_wprop3 (obj, "data", str);
      edma_wprop3 (obj, "font", font);
      edma_wprop3 (obj, "style", style);
    }
  else
    return -1;


  return 0;
}


ESint32 EDMAPROC 
GL_ITEM_FREE_TEXTdestroy (OBJID obj)
{
  GL_ITEM_FREE_TEXTData *m = edma_get_data_ref (obj);

  return 0;
}


ESint32 EDMAPROC 
GL_ITEM_FREE_TEXTrip (OBJID obj)
{
  GL_ITEM_FREE_TEXTData *m = edma_get_data_ref (obj);

  return 0;
}


ESint32 EDMAPROC 
GL_ITEM_FREE_TEXTrender (OBJID obj)
{
  ESint32               i, len;
  OBJID                 font;
  EDMAT_BUFFER          tex_buf;
  GLuint                list_base;
  TEX_FONT              *tex;
  /*****************************************************/
  float modelview_matrix[16];	
  GL_ITEM_FREE_TEXTData *m = edma_get_data_ref (obj);


  edma_met3 (obj, "SUPER>render");
  if (strlen (m->data) == 0) return 0;

  len = strlen (m->data);
  edma_rprop3 (obj, "font", &font);
  edma_rprop3 (font, "tex", &tex_buf);

  tex = (TEX_FONT *) tex_buf.dat;

#if 1
  glPushAttrib(GL_LIST_BIT | GL_CURRENT_BIT  | GL_ENABLE_BIT | GL_TRANSFORM_BIT);	
  glMatrixMode(GL_MODELVIEW);
  glDisable(GL_LIGHTING);
  glEnable(GL_TEXTURE_2D);
  glDisable(GL_DEPTH_TEST);
  glEnable(GL_BLEND);
  //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);	
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);	
  
  glGetFloatv(GL_MODELVIEW_MATRIX, modelview_matrix);
  
  glPushMatrix();
  glLoadIdentity();
  //glTranslatef(50,50,0);
  glMultMatrixf(modelview_matrix);
  
  //glColor4f (1.0, 1.0, 1.0, 1.0);
#endif

  glListBase (tex->list_base);  

  glCallLists(len, GL_UNSIGNED_BYTE, m->data);

#if 0
  edma_met3 (obj, "SUPER>render");  
  glTranslatef(-1.5,-1.5,0);
  glColor4f (1.0, 1.0, 1.0, 0.2);
  glCallLists(len, GL_UNSIGNED_BYTE, m->data);


  edma_met3 (obj, "SUPER>render");  
  glTranslatef(1.5,1.5,0);
  glColor4f (1.0, 1.0, 1.0, 0.2);
  glCallLists(len, GL_UNSIGNED_BYTE, m->data);

  edma_met3 (obj, "SUPER>render");  
  glTranslatef(1.5,-1.5,0);
  glColor4f (1.0, 1.0, 1.0, 0.2);
  glCallLists(len, GL_UNSIGNED_BYTE, m->data);

  edma_met3 (obj, "SUPER>render");  
  glTranslatef(0.0,-1.5,0);
  glColor4f (3.0, 1.0, 1.0, 0.2);
  glCallLists(len, GL_UNSIGNED_BYTE, m->data);

  edma_met3 (obj, "SUPER>render");  
  glTranslatef(5.0,-1.5,0);
  glColor4f (1.0, 1.0, 1.0, 0.2);
  glCallLists(len, GL_UNSIGNED_BYTE, m->data);

  edma_met3 (obj, "SUPER>render");  
  glTranslatef(-1.5,1.5,0);
  glColor4f (1.0, 1.0, 1.0, 0.2);
  glCallLists(len, GL_UNSIGNED_BYTE, m->data);

  /* Outline Color font */
  /*
  edma_met3 (obj, "SUPER>render");  
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);	
  glColor4f (0.0, 0.0, 0.0, 1);
  glCallLists(len, GL_UNSIGNED_BYTE, m->data);
  */

  edma_met3 (obj, "SUPER>render");  
  glBlendFunc(GL_SRC_ALPHA, GL_SRC_ALPHA_SATURATE);	
  glColor4f (0.0, 0.0, 0.0, 1);
  glCallLists(len, GL_UNSIGNED_BYTE, m->data);
#endif  
#if 0
  glPopMatrix();
  glPopAttrib();		
  
  pop_projection_matrix();

#endif  

  return 0;
}

ESint32 EDMAPROC 
GL_ITEM_FREE_TEXTrender_txt (CLASSID cid, EPChar style, EReal64 x, EReal64 y, EPChar txt)
{
  ESint32               i, len;
  OBJID                 font, styles;
  EDMAT_BUFFER          tex_buf;
  GLuint                list_base;
  TEX_FONT              *tex;
  EReal64               scale;
  /*****************************************************/
  float modelview_matrix[16];	

  len = strlen (txt);

  styles = (OBJID) edma_smet3 ("SINGLETON", "Instance", "STYLES");

  edma_met3 (styles, "get_data", style, &font, &scale);
  edma_rprop3 (font, "tex", &tex_buf);

  tex = (TEX_FONT *) tex_buf.dat;

  glLoadIdentity ();
  glTranslatef (x, y, -10);

  glDisable(GL_LIGHTING);
  glEnable(GL_TEXTURE_2D);
  glDisable(GL_DEPTH_TEST);
  glEnable(GL_BLEND);
  //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);	
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);	


  scale /= 80;

  glScalef (scale, scale, scale);

  glColor4f (1.0, 1.0, 1.0, 1.0);

  glListBase (tex->list_base);  
  glCallLists(len, GL_UNSIGNED_BYTE, txt);


  return 0;
}


ESint32 EDMAPROC 
GL_ITEM_FREE_TEXTsave (OBJID obj, OBJID stream)
{
  EChar            buffer[1024];
  GL_ITEM_FREE_TEXTData *m = edma_get_data_ref (obj);

  edma_met3 (stream, "puts", "GL_ITEM_FREE_TEXT\n");
  /* Let superclass save common information */
  edma_met3 (obj, "SUPER>save", stream);

  /* Save font id */
  edma_met3 (stream, "puts", m->style);
  edma_met3 (stream, "puts", "\n");
  /* Save Text data */
  edma_met3 (stream, "puts", m->data);
  edma_met3 (stream, "puts", "\n");



  return 0;
}

ESint32 EDMAPROC 
GL_ITEM_FREE_TEXTload (OBJID obj, OBJID stream)
{
  EChar            style[1024];
  EChar            str[1024];
  ESint32          len;
  GL_ITEM_FREE_TEXTData *m = edma_get_data_ref (obj);

  len = 1024;
  edma_met3 (stream, "gets", style, &len);
  style[strlen(style) - 1] = 0; /* Chomp String */

  /* Read text style*/
  len = 1024;
  edma_met3 (stream, "gets", str, &len);
  str[strlen(str) - 1] = 0; /* Chomp String */

  edma_printf_obj (obj, "Style: '%s'  Data: '%s'", style, str);

  /* Create text texture */
  return edma_met3 (obj, "create", style, str);

}



/***********************************************************************************/
/***********************************************************************************/
/***********************************************************************************/
/***********************************************************************************/

inline int 
next_p2 ( int a )
{
  int rval=1;

  while (rval < a) rval <<= 1;
  return rval;
}

/* Helper functions */
int 
build_display_list_item ( FT_Face face, char ch, EUint32 list_base, EUint32 * tex_base ) 
{
  FT_Glyph         glyph;
  FT_BitmapGlyph   bitmap_glyph;
  FT_Bitmap        bitmap;
  int              width, height, i, j;
  float            x, y;
  GLubyte          *expanded_data;

  /* Load the Glyph for our character. */
  if (FT_Load_Glyph( face, FT_Get_Char_Index( face, ch ), FT_LOAD_DEFAULT ))
    {
      edma_printf ("ERROR: %s", "Load_GLyph");
      return -1;
    }

  /* Move the face's glyph into a Glyph object. */

  if(FT_Get_Glyph( face->glyph, &glyph ))
    {
      edma_printf ("ERROR: %s", "FT_Get_Glyph");
      return -1;
    }
  
  
  /* Convert the glyph to a bitmap. */
  FT_Glyph_To_Bitmap( &glyph, ft_render_mode_normal, 0, 1 );
  bitmap_glyph = (FT_BitmapGlyph)glyph;

  bitmap = bitmap_glyph->bitmap;

  /* Use our helper function to get the widths of
   * the bitmap data that we will need in order to create
   * our texture.
   */
  width = next_p2( bitmap.width );
  height = next_p2( bitmap.rows );

  /* Allocate memory for the texture data.*/
  if ((expanded_data = (GLubyte *) malloc ( 4 * width * height)) == NULL)
    {
      edma_printf ("ERROR: %s", "Cannot allocalte memory for character texture");
      /* FIXME: Cleanup memory */
      return -1;
    }

  /* Generate texture from bitmap info */
  for(j = 0; j < height; j++) 
    {
      for(i = 0; i < width; i++)
	{
	  expanded_data[4 * (i + j * width)]= 
	    (i >= bitmap.width || j >= bitmap.rows) ? 0 : 255;
	    //0 :  bitmap.buffer[i + bitmap.width * j];
	    expanded_data[4 * (i + j * width) + 1] = 
	      (i >= bitmap.width || j >= bitmap.rows) ? 0: 255;
	    //0 : bitmap.buffer[i + bitmap.width * j];
	    expanded_data[4 * (i + j * width) + 2] = 
	      (i >= bitmap.width || j >= bitmap.rows) ? 0: 255;
	      //0 : bitmap.buffer[i + bitmap.width * j];
	    expanded_data[4 * (i + j * width) + 3] = 
	      (i >= bitmap.width || j >= bitmap.rows) ? //0: 1;
	      0 : bitmap.buffer[i + bitmap.width * j];
	}
    }


  glBindTexture   (GL_TEXTURE_2D, tex_base[(int)ch]);
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

  glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, width, height,
		0, GL_RGBA, GL_UNSIGNED_BYTE, expanded_data );
		//0, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, expanded_data );

  /* Free main memory */
  free (expanded_data);

  /* Generate display list */
  glNewList (list_base + ch, GL_COMPILE);

  glBindTexture (GL_TEXTURE_2D, tex_base[(int)ch]);

  glPushMatrix ();
  glTranslatef (0, bitmap_glyph->top - bitmap.rows, 0);


  x = (float) bitmap.width / (float) width;
  y = (float) bitmap.rows / (float) height;

  /* Draw character to effectively add it to display list */
  glBegin (GL_QUADS);

  glTexCoord2d (0, 0); glVertex2f (0,            bitmap.rows);
  glTexCoord2d (0, y); glVertex2f (0,            0);
  glTexCoord2d (x, y); glVertex2f (bitmap.width, 0);
  glTexCoord2d (x, 0); glVertex2f (bitmap.width, bitmap.rows);

  glEnd ();

  glPopMatrix ();
  glTranslatef (face->glyph->advance.x  >> 6 , 0,0);

  /* Display list done */
  glEndList();

  return 0;
}

