/*
 * Priscila: A GNU/EDMA OpenGL Presentation Tool
 * Copyright (C) 2004  David Martnez Oliveira
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 
/*
 * GL_SLIDE
 * Encapsulation of presentation slide
 * ---------------------------------------------------------
 * September 10th, 2004
 * - File Creation. GNU/EDMA in-line class
 * ----------------------------------------------------------
 * September, 11th, 2004
 * - Adding navigation information to slide files
 * - Updating navigation process function to navigate presentation
 * ------------------------------------------------------------------
 * September, 12th, 2004
 * - Added support to read a separated file with the FX for the slide
 * ---------------------------------------------------------------------
 * October, 5th, 2004
 * - Event 'process' method make virtual and overriden on mode change
 * - Added TEXT_EDIT_MODE and a simple cursor for Text editing
 * - Added character translation table to process SHIFT+KEY keystrokes
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include <SDL.h>
#include <SDL_ttf.h>

#include <edma.h>

#define MAX_ITEMS     50
#define MAX_FX        50
#define MAX_FONTS     20

#define VERY_FINE_STEP 0.001f
#define FINE_STEP     0.1f
#define COARSE_STEP   1.0f

#define NAVIGATION_MODE   0
#define EDIT_MODE         1
#define TEXT_EDIT_MODE    2

/* Character translation tables */
static int     ch_trans[255][3];

typedef struct gl_slide_t
{
  EPChar        filename;
  EDMAT_BUFFER  item;
  EDMAT_BUFFER  fx;
  EPChar        next;
  EPChar        prev;

  ESint32       mode;
  ESint32       current;
  EReal64       ed_incr;

  EPChar        current_error;
} GL_SLIDE_Data;


/* Method prototypes */
ESint32 EDMAPROC GL_SLIDErender (OBJID obj);
ESint32 EDMAPROC GL_SLIDEprocess (OBJID obj, EPVoid p);
ESint32 EDMAPROC GL_SLIDEload (OBJID obj, EPChar file_name);
ESint32 EDMAPROC GL_SLIDEsave (OBJID obj, EPChar file_name);
ESint32 EDMAPROC GL_SLIDEborn (OBJID obj);
ESint32 EDMAPROC GL_SLIDErip (OBJID obj);
ESint32 EDMAPROC GL_SLIDEfree_items (OBJID obj);

/* Helper functions */
int nav_process_key (OBJID obj, SDL_Event *event);
int edit_process_key (OBJID obj, SDL_Event *event);
int text_edit_process_key (OBJID obj, SDL_Event *event);

int
gl_slide_class_factory ()
{
  ESint32    i;
  CLASSID    cid;
  EChar      other_chars[] = "1234567890'<,. \0";
  EChar      shift_chars[] = "!\"$%&/()=?>;:_ \0";

  cid = edma_idf_get_free_class_id (EDMA_LOCAL_CLASS);

  edma_idf_set_class_name (cid, "GL_SLIDE");
  edma_idf_set_class_namespace (cid, "PRISCILA");
  edma_idf_set_class_version (cid, 0, 0);

  /* General Slide information */
  edma_add_local_class_property (cid, "filename", DT_EZSTRING, E_L, 0);
  edma_add_local_class_property (cid, "item", DT_EBUFFER, E_L, 0);
  edma_add_local_class_property (cid, "fx", DT_EBUFFER, E_L, 0);
  edma_add_local_class_property (cid, "next", DT_EZSTRING, E_L, 0);
  edma_add_local_class_property (cid, "prev", DT_EZSTRING, E_L, 0);

  /* EDIT Mode related information*/
  edma_add_local_class_property (cid, "mode", DT_ESINT32, E_L, 0);
  edma_add_local_class_property (cid, "current", DT_ESINT32, E_L, 0);
  edma_add_local_class_property (cid, "incr", DT_EREAL64, E_L, 0);

  edma_add_local_class_property (cid, "current_error", DT_EZSTRING, E_L, 0);


  /* GL_ITEM methods -> Virtual Method */
  edma_add_local_class_method (cid, "process", "P", (PPROC) nav_process_key, 1, 0, 0);

  edma_add_local_class_method (cid, "render", "", (PPROC) GL_SLIDErender, 0, 0, 0);
  edma_add_local_class_method (cid, "save", "Z", (PPROC) GL_SLIDEsave, 0, 0, 0);
  edma_add_local_class_method (cid, "load", "Z", (PPROC) GL_SLIDEload, 0, 0, 0);

  edma_add_local_class_method (cid, "free_items", "", (PPROC) GL_SLIDEfree_items, 0, 0, 0);
  edma_add_local_class_method (cid, "born", "", (PPROC) GL_SLIDEborn, 0, 0, 0);
  edma_add_local_class_method (cid, "rip", "", (PPROC) GL_SLIDErip, 0, 0, 0);

  edma_idf_set_class_id (cid);

  /* Initializes character translation table */
  /* Insert letters */

  for (i = 0; i < 255; i++) ch_trans[i][0] = ch_trans[i][1] = ch_trans[i][2] = -1;

  for (i = 'a'; i <= 'z'; i++) 
    {
      ch_trans[i][0] = i;
      ch_trans[i][1] = toupper(i);
    }

  /* Insert Number and SHIFT+Number */
  for (i = 0; i < strlen (other_chars); i++)
    {
      ch_trans[other_chars[i]][0] = other_chars[i];
      ch_trans[other_chars[i]][1] = shift_chars[i];
    }

  /* AltGr Simbols */
  ch_trans[186][2] = '\\'; /* 186 --> */
  ch_trans['1'][2] = '|';
  ch_trans['2'][2] = '@';
  ch_trans['3'][2] = '#';

  return 0;
}


ESint32 EDMAPROC
GL_SLIDErender (OBJID obj)
{
  ESint32         i;
  OBJID           *item, *fx;
  EChar           edit_str[1024], data[1024];
  GL_SLIDE_Data   *m = edma_get_data_ref (obj);

  item = (OBJID *) m->item.dat;
  fx = (OBJID *) m->fx.dat;
#if 0
  printf ("Fx 0: %d\n", fx[0]);
  i = 0;
  while (fx[i] != -1)
    {
      edma_met3 (fx[i], "run", item);
      i++;
    }
#endif
  for (i = 0; fx[i] != -1; i++) edma_met3 (fx[i], "run", item);

  for (i = 0; item[i] != -1; i++)
    {
      edma_met3 (item[i], "render");
      if ((m->mode == TEXT_EDIT_MODE) && (m->current == i))
	{
	  /* Draw a Quad at the end of render*/

	  glBindTexture (GL_TEXTURE_2D, 0);
	  
	  glTranslatef (10.0f, -6.0f, 0.0f);
	  glScalef (12.0f, 4.0f, 1.0f);
	  glBegin(GL_QUADS);
	  
	  glNormal3f( 0.0f, 0.0f, 0.1f); /* Normal is Up*/
	  
	  //glColor4f ( 0.2f, 0.2f, 0.8f, 0.99f);
	  glColor4f ( 0.2f, 0.2f, 0.8f, 0.7f);
	  glVertex3f(  1.0f,  1.0f,  0.0f);
	  glVertex3f(  1.0f, -1.0f,  0.0f);
	  glVertex3f( -1.0f, -1.0f,  0.0f);
	  glVertex3f( -1.0f, 1.0f,  0.0f);
	  
	  glEnd();
	}
    }

  if (m->mode) /* On edit mode render edit information */
    {
#if 1
      edma_rprop3 (obj, "current_error", data);
      snprintf (edit_str, 1024, "ITEM: [%d] | STEP: [%f] | ERROR: '%s'", 
		(int)m->current, m->ed_incr, data);
#else
      snprintf (edit_str, 1024, "ITEM: [%d] | STEP: [%lf]", 
		m->current, m->ed_incr);
#endif
#if 0
      /* Draw a Quad for text background */
      glBindTexture (GL_TEXTURE_2D, 0);

      glLoadIdentity();
      glTranslatef (-3.5f, -3.9f, -10.0f);
      glScalef (10.0f, 0.5f, 1.0f);
      glBegin(GL_QUADS);
      
      glNormal3f( 0.0f, 0.0f, 0.1f); /* Normal is Up*/
      
      glColor4f ( 0.2f, 0.2f, 0.8f, 0.99f);
      glVertex3f(  1.0f,  1.0f,  0.0f);
      glVertex3f(  1.0f, -1.0f,  0.0f);
      glVertex3f( -1.0f, -1.0f,  0.0f);
      glVertex3f( -1.0f, 1.0f,  0.0f);
	  
      glEnd();
#endif
      
      edma_smet3 ("GL_ITEM_FREE_TEXT", "render_txt", "SMALL1", -5.5, -3.9, edit_str);
      edma_rprop3 (item[m->current], "data", data);
      edma_smet3 ("GL_ITEM_FREE_TEXT", "render_txt", "SMALL1", -5.5, -3.7, data);


    }

  return 0;
}

ESint32 EDMAPROC 
GL_SLIDEload (OBJID obj, EPChar fname)
{
  OBJID           *item, *fx;
  OBJID           s, temp;
  ESint32         i, len;
  EChar           buffer[1024];
  EChar           filename[1024];
  GL_SLIDE_Data   *m = edma_get_data_ref (obj);

  if (fname == NULL) return -1;

  strncpy (filename, fname, 1024);

  /* Load Slide items */
  item = (OBJID *) m->item.dat;

  if ((s = edma_new_obj ("NFILE")) == -1)
    {
      edma_printf_err ("%s", "[load_slide] Cannot create STREAM object");
      return -1;
    }
  if ((edma_met3 (s, "open", filename, "rt")) == -1)
    {
      edma_printf_err ("%s", "[load_slide] Cannot open file for reading");
      return -1;
    }
  
  /* Read Navigation Information */
  i = 1024;
  edma_met3 (s, "gets", buffer, &i);
  buffer[strlen(buffer) - 1] = 0; /* Chomp */
  edma_wprop3 (obj, "next", buffer);

  i = 1024;
  edma_met3 (s, "gets", buffer, &i);
  buffer[strlen(buffer) - 1] = 0; /* Chomp */
  edma_wprop3 (obj, "prev", buffer);

  i = 0;
  /* Read Slide definition */
  while (! edma_met3 (s, "eof"))
    {
      edma_printf ("[LOAD] ===> Processing item %d", i);
      if (item[i] != -1) edma_free_obj (item[i]);
      if ((temp = edma_new_obj ("GL_ITEM")) == -1)
	{
	  edma_printf_err ("%s", "Cannot create GL_ITEM object... aborting");
	  return -1;
	}
      item[i] = edma_met3 (temp, "load", s);
      if (item[i] == -1) 
	{
	  edma_free_obj (temp);
	  break;
	}

      i++;
    }

  item[i] = -1;
  edma_met3 (s, "close");
  edma_free_obj (s);

  m->current = 0;
  edma_printf_obj (obj, "New Slide loaded '%s'", filename);
  edma_wprop3 (obj, "filename", filename);

  /* Load FX items */
  fx = (OBJID *) m->fx.dat;

  /* Initialise */
  fx[0] = -1;

  if ((s = edma_new_obj ("NFILE")) == -1)
    {
      edma_printf_err ("%s", "[load_slide] Cannot create STREAM object");
      return -1;
    }

  strncat (filename, ".fx", 1024);

  if ((edma_met3 (s, "open", filename, "rt")) == -1)
    {
      edma_printf_err ("%s", "[load_slide] No FX file found");
      /*edma_met3 (s, "close");*/
      edma_free_obj (s);
      return -1;
    }

  i = 0;
  /* Read Slide definition */
  while (! edma_met3 (s, "eof"))
    {
      edma_printf ("[LOAD] ===> Processing item %d", i);
      if (fx[i] != -1) edma_free_obj (fx[i]);
      len = 1024;
      edma_met3 (s, "gets", buffer, &len);  /* Load FX Class Name*/
      buffer[strlen(buffer) - 1] = 0;       /* Chomp string */
      if ((fx[i] = edma_new_obj (buffer)) == -1)
	{
	  edma_printf_err ("%s", "Cannot create FX object... aborting");
	  edma_met3 (s, "close");
	  edma_free_obj (s);
	  return -1;
	}
      fx[i] = edma_met3 (fx[i], "load", s);
      if (fx[i] == -1) break;
      i++;
    }

  fx[i] = -1;
  fx[i + 1] = -1;
  edma_met3 (s, "close");
  edma_free_obj (s);

  return 0;
}

ESint32 EDMAPROC 
GL_SLIDEsave (OBJID obj, EPChar filename)
{
  OBJID           *item;
  OBJID           s;
  ESint32         i;
  GL_SLIDE_Data   *m = edma_get_data_ref (obj);

  if (filename == NULL) return -1;

  item = (OBJID *) m->item.dat;

  if ((s = edma_new_obj ("NFILE")) == -1)
    {
      edma_printf_err ("%s", "[save_slide] Cannot create STREAM object");
      return -1;
    }
  if ((edma_met3 (s, "open", filename, "wt")) == -1)
    {
      edma_printf_err ("%s", "[save_slide] Cannot open file for writting");
      return -1;
    }

  edma_printf_obj (obj, "next: '%s' prev: '%s'", m->next, m->prev);
  /* Save Navigation Information */
  edma_met3 (s, "puts", m->next, 1024);
  edma_met3 (s, "puts", "\n");
  edma_met3 (s, "puts", m->prev, 1024);
  edma_met3 (s, "puts", "\n");


  for (i = 0; item[i] != -1; i++)
    {
      edma_printf_obj (obj, "Saving item %d", i);
      edma_met3 (item[i], "save", s);
    }

  edma_met3 (s, "puts", "\n", 1024); /* EOF Separator */
  edma_met3 (s, "close");
  edma_free_obj (s);

  return 0;
}

ESint32 EDMAPROC 
GL_SLIDEborn (OBJID obj)
{
  ESint32         i;
  OBJID           *item; //, *fx;
  GL_SLIDE_Data   *m = edma_get_data_ref (obj);

  /* Initialization */
  /* Allocate slide items array */
  if ((edma_buffer_alloc (&m->item, sizeof(OBJID) * MAX_ITEMS)) < 0)
    {
      edma_printf_obj (obj, "%s", "ERROR: Cannot allocate memory for slide items");
      return -1;
    }

  item = (OBJID *) m->item.dat;
  for (i = 0; i < MAX_ITEMS; i++) item[i] = -1;

  /* Allocate slide fx array */
  if ((edma_buffer_alloc (&m->fx, sizeof(OBJID) * MAX_FX)) < 0)
    {
      edma_printf_obj (obj, "%s", "ERROR: Cannot allocate memory for slide items");
      return -1;
    }

  item = (OBJID *) m->fx.dat;
  for (i = 0; i < MAX_FX; i++) item[i] = -1;

  m->mode = NAVIGATION_MODE;
  m->ed_incr = FINE_STEP;
  m->current = 0;
  edma_wprop3 (obj, "current_error", "OK");

  return 0;
}

ESint32 EDMAPROC 
GL_SLIDErip (OBJID obj)
{
  ESint32         i;
  OBJID           *item;
  GL_SLIDE_Data   *m = edma_get_data_ref (obj);

  item = (OBJID *) m->item.dat;
  /* Destroy all slide items */
  for (i = 0; i < MAX_ITEMS && item[i] != -1; i++) edma_free_obj (item[i]);

  item = (OBJID *) m->fx.dat;
  /* Destroy all slide items */
  for (i = 0; i < MAX_FX && item[i] != -1; i++) edma_free_obj (item[i]);

  return 0;
}

ESint32 EDMAPROC 
GL_SLIDEfree_items (OBJID obj)
{
  ESint32         i;
  OBJID           *item;
  GL_SLIDE_Data   *m = edma_get_data_ref (obj);

  edma_printf_obj (obj, "%s", "Freeing slide resources");
  item = (OBJID *) m->item.dat;
  /* Destroy all slide items */
  for (i = 0; i < MAX_ITEMS && item[i] != -1; i++) 
    {
      edma_free_obj (item[i]);
      item[i] = -1;

    }

  item = (OBJID *) m->fx.dat;
  /* Destroy all slide items */
  for (i = 0; i < MAX_FX && item[i] != -1; i++) 
    {
      edma_free_obj (item[i]);
      item[i] = -1;

    }

  return 0;
}


/* Helper functions */
/* Key action processing on TEXT EDIT Mode*/
int
text_edit_process_key (OBJID obj, SDL_Event *event)
{
  OBJID           *item;
  ESint32         len, ch, mod, mod1;
  EChar           txt[1024];   /* Buffer for text editing */
  GL_SLIDE_Data   *m = edma_get_data_ref (obj);
  
  item = (OBJID *) m->item.dat;

  //edma_wprop3 (obj, "current_error", "OK");
  
  /* Just process ESC key. 
   * Edit mode keys requires additional methods in GL_ITEM class
   */
  switch (event->key.keysym.sym)
    {
    case SDLK_ESCAPE:  /* Finish application */
      return 1;
    case SDLK_RETURN:
      {
	m->mode = EDIT_MODE;
	edma_over_met (obj, "process", edma_get_obj_pobj (obj), (PPROC) edit_process_key);
	return 0;
      }
    case SDLK_BACKSPACE:
      {
	edma_rprop3 (item[m->current], "data", txt);
	len = strlen (txt);
	if (len == 0)
	  {
	    txt[0] = 0;
	    edma_wprop3 (item[m->current], "data", txt);
	    return 0;
	  }
	txt[len - 1] = 0;
	edma_wprop3 (item[m->current], "data", txt);
	return 0;
      }
    default:
      /* Do nothing*/
    }
  ch = event->key.keysym.sym;

  if (ch >= 32 && ch < SDLK_NUMLOCK)
    {

      edma_rprop3 (item[m->current], "data", txt);

      mod1 = event->key.keysym.mod;
      mod = 0;
      if ((mod1 & KMOD_SHIFT) != 0L)
	mod = 1;

      if ((mod1 & KMOD_MODE) != 0L)
	  mod = 2;

      edma_printf ("Key Sym: %d:'%c' Key Mod: %x MyMod: %d", ch, ch, event->key.keysym.mod, mod);
      if ((ch = ch_trans[ch][mod]) == -1)
	return 0;

      len = strlen(txt);

      txt[len] = ch;

      txt[len + 1] = 0;
      edma_wprop3 (item[m->current], "data", txt);
    }

  return 0;
}

/* Key action processing on NAVIGATION Mode */
int
nav_process_key (OBJID obj, SDL_Event *event)
{
  GL_SLIDE_Data   *m = edma_get_data_ref (obj);
  
  edma_wprop3 (obj, "current_error", "OK");

  /* Just process ESC key. 
   * Edit mode keys requires additional methods in GL_ITEM class
   */
  switch (event->key.keysym.sym)
    {
    case SDLK_ESCAPE:  /* Finish application */
      return 1;
    case SDLK_RIGHT:
      {
	/* Go to next slide */ 
	if (strcmp (m->next, "NONE") == 0)
	  return 0;
	edma_printf ("%s", "Moving to next slide in presentation");

	edma_met3 (obj, "free_items");
	edma_met3 (obj, "load", m->next);
	return 0;
      }
    case SDLK_LEFT:
      {
	/* Go to previous slide */
	if (strcmp (m->prev, "NONE") == 0)
	  return 0;
	edma_printf ("%s", "Moving to previous slide in presentation");

	edma_met3 (obj, "free_items");
	edma_met3 (obj, "load", m->prev);

	return 0;
      }
    case SDLK_INSERT:
      {
	/* Enter Edit Mode */
	edma_printf ("%s", "Entering EDIT Mode");
	m->mode = EDIT_MODE;
	/* Overwrites process method to use edit_process_key function */
	edma_over_met (obj, "process", edma_get_obj_pobj (obj), (PPROC) edit_process_key);
	return 0;
      }     
    default:
      return 0;
    }
  return 0;
}

/* Key action processing on EDIT Mode */
int
edit_process_key (OBJID obj, SDL_Event *event)
{
  OBJID           *item, aux;
  EReal64         x, y, z, alpha;
  ESint32         i;
  GL_SLIDE_Data   *m = edma_get_data_ref (obj);

  item = (OBJID *) m->item.dat;

  edma_wprop3 (obj, "current_error", "OK");

  /* Just process ESC key. 
   * Edit mode keys requires additional methods in GL_ITEM class
   */
  switch (event->key.keysym.sym)
    {
    case SDLK_ESCAPE:  /* Finish application */
      return 1;

    case SDLK_0:
      {
	m->ed_incr = VERY_FINE_STEP;
	return 0;
      }

    case SDLK_1:
      {
	m->ed_incr = FINE_STEP;
	return 0;
      }
    case SDLK_2:
      {
	m->ed_incr = COARSE_STEP;
	return 0;
      }
    case SDLK_RIGHT:
      {
	edma_met3 (item[m->current], "get_pos", &x, &y, &z);
	x += m->ed_incr;
	edma_met3 (item[m->current], "set_pos", x, y, z);
	return 0;
      }
    case SDLK_LEFT:
      {
	edma_met3 (item[m->current], "get_pos", &x, &y, &z);
	x -= m->ed_incr;
	edma_met3 (item[m->current], "set_pos", x, y, z);
	return 0;
      }
    case SDLK_DOWN:
      {
	edma_met3 (item[m->current], "get_pos", &x, &y, &z);
	y -= m->ed_incr;
	edma_met3 (item[m->current], "set_pos", x, y, z);
	return 0;
      }
    case SDLK_UP:
      {
	edma_met3 (item[m->current], "get_pos", &x, &y, &z);
	y += m->ed_incr;
	edma_met3 (item[m->current], "set_pos", x, y, z);
	return 0;
      }
    case SDLK_a:
      {
	edma_met3 (item[m->current], "get_scale", &x, &y, &z);
	x += m->ed_incr;
	y += m->ed_incr;
	z += m->ed_incr;
	edma_met3 (item[m->current], "set_scale", x, y, z);
	return 0;
      }

    case SDLK_z:
      {
	edma_met3 (item[m->current], "get_scale", &x, &y, &z);
	x -= m->ed_incr;
	y -= m->ed_incr;
	z -= m->ed_incr;
	edma_met3 (item[m->current], "set_scale", x, y, z);
	return 0;
      }

    case SDLK_b:
      {
	edma_met3 (item[m->current], "get_rot", &x, &y, &z);
	x -= m->ed_incr;
	edma_met3 (item[m->current], "set_rot", x, y, z);
	return 0;
      }

    case SDLK_h:
      {
	edma_met3 (item[m->current], "get_rot", &x, &y, &z);
	x += m->ed_incr;
	edma_met3 (item[m->current], "set_rot", x, y, z);
	return 0;
      }

    case SDLK_n:
      {
	edma_met3 (item[m->current], "get_rot", &x, &y, &z);
	y -= m->ed_incr;
	edma_met3 (item[m->current], "set_rot", x, y, z);
	return 0;
      }

    case SDLK_j:
      {
	edma_met3 (item[m->current], "get_rot", &x, &y, &z);
	y += m->ed_incr;
	edma_met3 (item[m->current], "set_rot", x, y, z);
	return 0;
      }

    case SDLK_m:
      {
	edma_met3 (item[m->current], "get_rot", &x, &y, &z);
	z -= m->ed_incr;
	edma_met3 (item[m->current], "set_rot", x, y, z);
	return 0;
      }

    case SDLK_k:
      {
	edma_met3 (item[m->current], "get_rot", &x, &y, &z);
	z += m->ed_incr;
	edma_met3 (item[m->current], "set_rot", x, y, z);
	return 0;
      }

    case SDLK_q:
      {
	edma_met3 (item[m->current], "get_color", &x, &y, &z, &alpha);
	alpha -= m->ed_incr;
	edma_met3 (item[m->current], "set_color", x, y, z, alpha);
	return 0;
      }

    case SDLK_w:
      {
	edma_met3 (item[m->current], "get_color", &x, &y, &z, &alpha);
	alpha += m->ed_incr;
	edma_met3 (item[m->current], "set_color", x, y, z, alpha);
	return 0;
      }

    case SDLK_TAB:
      {
	m->current ++;
	if (item[m->current] == -1)
	  m->current = 0;
	return 0;
      }

    case SDLK_PAGEUP:
      {
	if (((m->current + 1) < MAX_ITEMS) && (item[m->current + 1]) == -1) 
	  return 0;
	else
	  {
	    aux = item[m->current + 1];
	    item[m->current + 1] = item[m->current];
	    item[m->current] = aux;
	    m->current++;
	    return 0;
	  }
      }

    case SDLK_PAGEDOWN:
      {
	if (m->current == 0)
	  return 0;
	else
	  {
	    aux = item[m->current - 1];
	    item[m->current - 1] = item[m->current];
	    item[m->current] = aux;
	    m->current--;
	    return 0;
	  }
      }

    case SDLK_t:
      {
	/* Locate an available position in item list */
	for (i = 0; i < MAX_ITEMS; i++)
	  if (item[i] == -1)
	    break;
	if (i >= MAX_ITEMS)
	  return 0;

	/* Insert new item in list and initialize it */
	m->current = i;
	item[m->current + 1] = -1;

	if ((item[m->current] = edma_new_obj ("GL_ITEM_FREE_TEXT")) == -1)
	  return 1;

	edma_met3 (item[m->current], "create", "NORMAL1", "Write Here");
	edma_met3 (item[m->current], "set_pos", 0.0, 0.0, -30.0);
	edma_met3 (item[m->current], "set_scale", 0.05, 0.05, 0.05);
	edma_met3 (item[m->current], "set_color", 1.0, 1.0, 1.0, 1.0);

	/* Setup Text edit mode and new key event call-back*/
	m->mode = TEXT_EDIT_MODE;
	edma_over_met (obj, "process", edma_get_obj_pobj (obj), (PPROC) text_edit_process_key);

	return 0;
      }

    case SDLK_DELETE:
      {
	if (item[m->current] != -1) 
	  {
	    
	    /* Locate an available position in item list */
	    for (i = 0; i < MAX_ITEMS; i++)
	      if (item[i] == -1)
		break;
	    if (i >= MAX_ITEMS)
	      return 0;

	    edma_printf ("%d item in current slide. Removing item %d", i, m->current);
	    edma_free_obj (item[m->current]);
	    //item[m->current] = -1;
	    i--;
	    if (i < 0) 
	      {
		m->current = 0;
		return 0;
	      }
	    edma_printf ("Current Item is now %d. Size is %d", m->current, i);
	    memmove (item + m->current, item + m->current + 1, (i - m->current) *sizeof(OBJID*));
	    if (m->current == i)
	      m->current --;
	    item[i] = -1;
	  }

	return 0;
      }

    case SDLK_RETURN:
      {
	if ((aux = edma_cast_obj (item[m->current], "GL_ITEM_FREE_TEXT")) == -1)
	  {
	    edma_wprop3 (obj, "current_error", "Non Editable Object");
	    return 0;
	  }
	else
	  {
	    m->mode = TEXT_EDIT_MODE;
	    edma_over_met (obj, "process", edma_get_obj_pobj (obj), 
			   (PPROC) text_edit_process_key); 
	  }
	return 0;
      }

    case SDLK_INSERT:
      {
	/* Enter Edit Mode */
	edma_printf ("%s", "Entering NAVIGATION Mode");
	m->mode = NAVIGATION_MODE;

	/* Restore original Navigation process key method */
	edma_restore_met (obj, "process");
	return 0;
      }

      /* Load/Save function only available in EDIT Mode*/
    case SDLK_s:
      {
	edma_printf ("Saving slide on file %s", m->filename);
	edma_met3 (obj, "save", m->filename);
	return 0;
      }
    case SDLK_l:
      {
	edma_printf ("Loading slide from file %s", m->filename);
	edma_met3 (obj, "load", m->filename);
	return 0;
      }
    default:
      return 0;
    }
  return 0;
}



