/*
 * Helper functions for OpenGL & SDL
 */

#include <stdio.h>
#include <stdlib.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include <SDL.h>
#include <SDL_ttf.h>
#include <SDL_image.h>

#include "sdlgl_tex.h"

#define SDLGL_FONT_C

/* Internal datatypes */

typedef struct sdlgl_font_t
{
  TTF_Font    *ttf_font;
  //SDLGL_TEX   mask_tex;
  //SDLGL_TEX   txt_tex;
  int         w,h;
  //GLfloat     scale;
  int         bg_r, bg_b, bg_g;
  int         fg_r, fg_b, fg_g;
  int         style;
} *SDLGL_FONT;

typedef struct sdlgl_txt_t
{
  SDLGL_FONT  font;
  char        *txt;
  int         w,h;
  SDLGL_TEX   mask_tex;
  SDLGL_TEX   txt_tex;
  GLfloat     scale;
} *SDLGL_TXT;

#include "sdlgl_font.h"

SDLGL_FONT
SDLGL_font_new (char *font_path, unsigned int size)
{
  SDLGL_FONT  the_font;

  if ((the_font = (SDLGL_FONT) malloc (sizeof (struct sdlgl_font_t))) == NULL)
    {
      fprintf (stderr, "Can't alloc memory for SDLGL_FONT object.\n");
      return NULL;
    }
  memset (the_font, 0, sizeof(struct sdlgl_font_t));
  if ((the_font->ttf_font = TTF_OpenFont (font_path, size)) == NULL)
    {
      fprintf (stderr, "Can't load font '%s' of size %d\n", font_path, size);
      free (the_font);
      return NULL;
    }
  //the_font->mask_tex = SDLGL_new_texture ();
  //the_font->txt_tex = SDLGL_new_texture();
  //the_font->scale = 0.5;
  the_font->bg_r = the_font->bg_g = the_font->bg_b = 0;
  the_font->fg_r = the_font->fg_g = the_font->fg_b = 0;
  the_font->fg_r = 255;
  return the_font;
}

int
SDLGL_font_free (SDLGL_FONT font)
{
  //SDLGL_free_texture (font->txt_tex);
  //SDLGL_free_texture (font->mask_tex);
  TTF_CloseFont (font->ttf_font);
  free (font);
  return 0;
}

int SDLGL_font_set_fgcolor (SDLGL_FONT font, int r, int g, int b)
{
  font->fg_r = r;
  font->fg_g = g;
  font->fg_b = b;
  return 0;
}


int
SDLGL_font_build_tex (SDLGL_FONT font, char *str, SDLGL_TEX *mask, SDLGL_TEX *txt)
{
  SDL_Color   txt_color, bg_color, mask_color;
  SDL_Surface *the_txt, *txt_mask;

  if (!font)
    {
      fprintf (stderr, "[%s] Invalid font object\n", __FUNCTION__);
      return -1;
    }

  txt_color.r = font->fg_r;
  txt_color.g = font->fg_g;
  txt_color.b = font->fg_b;

  bg_color.r = font->bg_r;
  bg_color.b = font->bg_b;
  bg_color.g = font->bg_g;

  mask_color.r = 255;
  mask_color.b = 255;
  mask_color.g = 255;

  printf ("[%s] Rendeing text image for text '%s' with font %p\n", __FUNCTION__, str, font);
  if ((the_txt = TTF_RenderText_Shaded (font->ttf_font, str, txt_color, bg_color)) == NULL)
    {
      fprintf (stderr, "[%s] Can't render text surface\n", __FUNCTION__);
      return -1;
    }
  printf ("[%s] New texture created of size (%d, %d)\n", __FUNCTION__, the_txt->w, the_txt->h);
  printf ("[%s] Rendeing text mask for text '%s' with font %p\n", __FUNCTION__, str, font);
  if ((txt_mask = TTF_RenderText_Shaded (font->ttf_font, str, bg_color, mask_color)) == NULL)
    {
      fprintf (stderr, "[%s] Can't render mask surface\n", __FUNCTION__);
      SDL_FreeSurface (the_txt);
      return -1;
    }

  /* FIXME: Use accessor function for this information or rely in SDLGL_TEX objects
  *         created down*/
  font->w = the_txt->w;
  font->h = the_txt->h;
  
  printf ("[%s] Creating SDLGL_TEX objects\n", __FUNCTION__);
  /* Create SDLGL_TEX objects */
  if ((*mask = SDLGL_new_texture ()) == NULL)
    {
      fprintf (stderr, "[%s] Can't create SDLGL_TEX object for text mask\n", __FUNCTION__);
      SDL_FreeSurface (the_txt);
      SDL_FreeSurface (txt_mask);
      return -1;
    }
  if ((*txt = SDLGL_new_texture()) == NULL)
    {
      fprintf (stderr, "[%s] Can't create SDLGL_TEX object for text\n", __FUNCTION__);
      SDL_FreeSurface (the_txt);
      SDL_FreeSurface (txt_mask);
      SDLGL_free_texture (*mask);
      return -1;      
    }

  printf ("[%s] Building GL textures for the text\n", __FUNCTION__);
  /* Create GL texture from SDL surface*/
  if ((SDLGL_read_from_surface (*mask, the_txt)) == -1)
    {
      fprintf (stderr, "[%s] Can't create GL texture\n", __FUNCTION__);
      return -1;
    }
  if ((SDLGL_read_from_surface (*txt, txt_mask)) == -1)
    {
      fprintf (stderr, "[%s] Can't create GL texture\n", __FUNCTION__);
      return -1;      
    }

  printf ("[%s] Done\n", __FUNCTION__);
  return 0;
}

int 
SDLGL_font_simple_render (SDLGL_FONT font)
{
#if 0
  GLfloat  txt_texMaxX, txt_texMaxY, txt_texMinX, txt_texMinY;
  GLfloat  x0, y0, x1, y1;
  GLfloat  txt_texcoord[4];

  /* Push user tranformation before calling this function ????*/
  SDLGL_tex_get_map_coord (font->txt_tex, txt_texcoord);
  txt_texMaxX = txt_texcoord[0];
  txt_texMinY = txt_texcoord[1];
  txt_texMinX = txt_texcoord[2];
  txt_texMaxY = txt_texcoord[3];

  /* Bind  the Text Mask texture*/
  glBindTexture (GL_TEXTURE_2D, SDLGL_tex_get_gltex (font->mask_tex));


  /* Required to delete background image*/
  glEnable (GL_BLEND);
  glBlendFunc (GL_DST_COLOR,GL_ZERO);
  /* Correct QUAD scalling to maintain text aspect ratio*/
  glScalef (font->scale * ((float)font->w/(float)font->h), font->scale, font->scale);
  
  /* Just render a square QUAD */
  x0 = y0 = -1.0f;
  x1 = y1 =  1.0f;

  glBegin(GL_QUADS);
  
  glNormal3f  ( 0.0f, 0.0f, 1.0f); /* Normal is Up*/	
  glTexCoord2f( txt_texMaxX,  txt_texMinY); glVertex2i( x0,  y1);
  glTexCoord2f( txt_texMinX,  txt_texMinY); glVertex2i( x1,  y1);
  glTexCoord2f( txt_texMinX,  txt_texMaxY); glVertex2i( x1,  y0);
  glTexCoord2f( txt_texMaxX,  txt_texMaxY); glVertex2i( x0,  y0);

  glEnd();

  /* Solid text follows. We need to update the transformation code
   * to fit the mask just rendered above*/
  glBindTexture (GL_TEXTURE_2D, SDLGL_tex_get_gltex (font->txt_tex));

  glTranslatef (0.0f, 0.0f, 0.09f);
  //glEnable(GL_BLEND);
  glBlendFunc (GL_ONE,GL_ONE);

  //glScalef (font->scale * (font->w/font->h), font->scale, font->scale);

  x0 = y0 = -1;
  y1 = x1 =  1;
  glBegin(GL_QUADS);
  
  glNormal3f( 0.0f, 0.0f, 1.0f); /* Normal is Up*/	
  glTexCoord2f( txt_texMaxX,  txt_texMinY); glVertex2i( x0,  y1);
  glTexCoord2f( txt_texMinX,  txt_texMinY); glVertex2i( x1,  y1);
  glTexCoord2f( txt_texMinX,  txt_texMaxY); glVertex2i( x1,  y0);
  glTexCoord2f( txt_texMaxX,  txt_texMaxY); glVertex2i( x0,  y0);
  glEnd();

  glDisable(GL_BLEND);
  //glEnable(GL_DEPTH_TEST);
#else
  printf ("[%s] Not implemented\n", __FUNCTION__);
#endif
  return 0;
}

int
SDLGL_font_render_txt (SDLGL_FONT font, char *txt)
{
#if 0
  SDLGL_font_build_tex (font, txt);
  SDLGL_font_simple_render (font);
#endif
  return 0;
}

SDLGL_TXT
SDLGL_txt_new (SDLGL_FONT font, char *txt, GLfloat scale)
{
  SDLGL_TXT   the_txt;

  if (!font)
    return NULL;

  if ((the_txt = malloc (sizeof (struct sdlgl_txt_t))) == NULL)
    {
      fprintf (stderr, "Can't create SDLGL_TXT object\n");
      return NULL;
    }
  printf ("[%s] Allocated %d bytes, now initializating\n", __FUNCTION__, sizeof (struct sdlgl_txt_t));
  memset (the_txt, 0, sizeof (struct sdlgl_txt_t));
  the_txt->font = font;
  //the_txt->txt = strdup (txt);

  /* FIXME: Use SDLGL_FONT object to build textures */
  printf ("[%s] Building textures for string '%s'\n", __FUNCTION__, txt);
  SDLGL_font_build_tex (font, txt, &the_txt->mask_tex, &the_txt->txt_tex);
  printf ("[%s] Final Initializations\n", __FUNCTION__);
  the_txt->w = font->w;
  the_txt->h = font->h;
  the_txt->scale = scale;
  return the_txt;
}

int
SDLGL_txt_free (SDLGL_TXT txt)
{
  //free (txt->txt);

  SDLGL_free_texture (txt->mask_tex);
  SDLGL_free_texture (txt->txt_tex);
  free (txt);
  return 0;
}

GLfloat 
SDLGL_txt_get_aspect (SDLGL_TXT txt)
{
  return (txt->w/txt->h);
}

int 
SDLGL_txt_simple_render (SDLGL_TXT txt)
{
  GLfloat  txt_texMaxX, txt_texMaxY, txt_texMinX, txt_texMinY;
  GLfloat  x0, y0, x1, y1;
  GLfloat  txt_texcoord[4];

  /* Push user tranformation before calling this function ????*/
  SDLGL_tex_get_map_coord (txt->txt_tex, txt_texcoord);
  txt_texMaxX = txt_texcoord[0];
  txt_texMinY = txt_texcoord[1];
  txt_texMinX = txt_texcoord[2];
  txt_texMaxY = txt_texcoord[3];

  /* Bind  the Text Mask texture*/
  //glBindTexture (GL_TEXTURE_2D, SDLGL_tex_get_gltex (txt->mask_tex));
  glBindTexture (GL_TEXTURE_2D, SDLGL_tex_get_gltex (txt->txt_tex));

  /* Required to delete background image*/
  glEnable (GL_BLEND);
  glBlendFunc (GL_DST_COLOR,GL_ZERO);
  /* Correct QUAD scalling to maintain text aspect ratio*/
  glScalef (txt->scale * (txt->w/txt->h), txt->scale, txt->scale);
  
  /* Just render a square QUAD */
  x0 = y0 = -1.0f;
  x1 = y1 =  1.0f;

  glBegin(GL_QUADS);
  
  glNormal3f  ( 0.0f, 0.0f, 1.0f); /* Normal is Up*/	
  glTexCoord2f( txt_texMaxX,  txt_texMinY); glVertex2i( x0,  y1);
  glTexCoord2f( txt_texMinX,  txt_texMinY); glVertex2i( x1,  y1);
  glTexCoord2f( txt_texMinX,  txt_texMaxY); glVertex2i( x1,  y0);
  glTexCoord2f( txt_texMaxX,  txt_texMaxY); glVertex2i( x0,  y0);

  glEnd();

  /* Solid text follows. We need to update the transformation code
   * to fit the mask just rendered above*/
  //glBindTexture (GL_TEXTURE_2D, SDLGL_tex_get_gltex (txt->txt_tex));
  glBindTexture (GL_TEXTURE_2D, SDLGL_tex_get_gltex (txt->mask_tex));

  glTranslatef (0.0f, 0.0f, 0.09f);
  //glEnable(GL_BLEND);
  glBlendFunc (GL_ONE,GL_ONE);

  //glScalef (font->scale * (font->w/font->h), font->scale, font->scale);

  x0 = y0 = -1;
  y1 = x1 =  1;
  glBegin(GL_QUADS);
  
  glNormal3f( 0.0f, 0.0f, 1.0f); /* Normal is Up*/	
  glTexCoord2f( txt_texMaxX,  txt_texMinY); glVertex2i( x0,  y1);
  glTexCoord2f( txt_texMinX,  txt_texMinY); glVertex2i( x1,  y1);
  glTexCoord2f( txt_texMinX,  txt_texMaxY); glVertex2i( x1,  y0);
  glTexCoord2f( txt_texMaxX,  txt_texMaxY); glVertex2i( x0,  y0);
  glEnd();

  glDisable(GL_BLEND);
  //glEnable(GL_DEPTH_TEST);
  return 0;
}


int 
SDLGL_txt_simple_render1 (SDLGL_TXT txt)
{
  GLfloat  txt_texMaxX, txt_texMaxY, txt_texMinX, txt_texMinY;
  GLfloat  x0, y0, x1, y1;
  GLfloat  txt_texcoord[4];

  /* Push user tranformation before calling this function ????*/
  SDLGL_tex_get_map_coord (txt->txt_tex, txt_texcoord);
  txt_texMaxX = txt_texcoord[0];
  txt_texMinY = txt_texcoord[1];
  txt_texMinX = txt_texcoord[2];
  txt_texMaxY = txt_texcoord[3];

  /* Correct QUAD scalling to maintain text aspect ratio*/
  glScalef (txt->scale * (txt->w/txt->h), txt->scale, txt->scale);

#if 0
  /* Bind  the Text Mask texture*/
  //glBindTexture (GL_TEXTURE_2D, SDLGL_tex_get_gltex (txt->mask_tex));
  glBindTexture (GL_TEXTURE_2D, SDLGL_tex_get_gltex (txt->txt_tex));

  /* Required to delete background image*/
  glEnable (GL_BLEND);
  glBlendFunc (GL_DST_COLOR,GL_ZERO);

  
  /* Just render a square QUAD */
  x0 = y0 = -1.0f;
  x1 = y1 =  1.0f;

  glBegin(GL_QUADS);
  
  glNormal3f  ( 0.0f, 0.0f, 1.0f); /* Normal is Up*/	
  glTexCoord2f( txt_texMaxX,  txt_texMinY); glVertex2i( x0,  y1);
  glTexCoord2f( txt_texMinX,  txt_texMinY); glVertex2i( x1,  y1);
  glTexCoord2f( txt_texMinX,  txt_texMaxY); glVertex2i( x1,  y0);
  glTexCoord2f( txt_texMaxX,  txt_texMaxY); glVertex2i( x0,  y0);

  glEnd();
#endif
  /* Solid text follows. We need to update the transformation code
   * to fit the mask just rendered above*/
  //glBindTexture (GL_TEXTURE_2D, SDLGL_tex_get_gltex (txt->txt_tex));
  glBindTexture (GL_TEXTURE_2D, SDLGL_tex_get_gltex (txt->mask_tex));

  glTranslatef (0.0f, 0.0f, 0.09f);
  //glEnable(GL_BLEND);
  //glBlendFunc (GL_ONE,GL_ONE);

  //glScalef (font->scale * (font->w/font->h), font->scale, font->scale);

  x0 = y0 = -1;
  y1 = x1 =  1;
  glBegin(GL_QUADS);
  
  glNormal3f( 0.0f, 0.0f, 1.0f); /* Normal is Up*/	
  glTexCoord2f( txt_texMaxX,  txt_texMinY); glVertex2i( x0,  y1);
  glTexCoord2f( txt_texMinX,  txt_texMinY); glVertex2i( x1,  y1);
  glTexCoord2f( txt_texMinX,  txt_texMaxY); glVertex2i( x1,  y0);
  glTexCoord2f( txt_texMaxX,  txt_texMaxY); glVertex2i( x0,  y0);
  glEnd();

  //glDisable(GL_BLEND);
  //glEnable(GL_DEPTH_TEST);

  return 0;
}
